"""Community curated mappings between biomedical entities."""

from .graph import get_false_graph, get_predictions_graph, get_true_graph  # noqa:F401
from .resources import (  # noqa:F401
    SemanticMapping,
    load_false_mappings,
    load_mappings,
    load_mappings_subset,
    load_predictions,
    load_unsure,
)
